Generalizable All Variables.
Set Implicit Arguments.

(** * Classes for algebraic definitions using Coq equality *)

(** ** Unit *)
Class Unit `(op:A -> A -> A) (e:A) : Type := {
  left_unit : forall a, op e a = a;
  right_unit: forall a, op a e = a
}.

(** ** Associativity  *)
Class  Associative `(op:A -> A -> A) : Type :=
  associative : forall (x y z: A), op (op x y) z =  op x (op y z).

(** ** Monoid *)
Class Monoid `(op:A -> A -> A) (e:A) : Type := 
  {
    monoid_associative :> Associative op; 
    monoid_unit :> Unit op e
  }.

(** ** Instances *)

Require Import Arith.
Require Import Program.

(** *** (plus,0) is a monoid *)

Local Obligation Tactic := (program_simpl; auto with arith).

Program Instance associative_plus : Associative plus.
Program Instance unit_plus_O : Unit plus 0.
Program Instance monoid_plus_0 : Monoid plus 0.

(** *** (mult,1) is a monoid *)

Program Instance associative_mult : Associative mult.
Program Instance unit_mult_1 : Unit mult 1.
Program Instance monoid_mult_1 : Monoid mult 1.

(** *** (++,[]) is a monoid *)

Local Obligation Tactic := 
  (program_simpl; List.ssimpl_list; auto using List.app_assoc).

Program Instance associative_app {A:Type}: Associative (List.app(A:=A)).
Program Instance unit_app_nil {A:Type} : Unit (@List.app A) (@nil A).
Program Instance monoid_app_nil {A:Type} : Monoid (@List.app A) (@nil A).